package ch.hevs.isi;

import ch.hevs.isi.core.BooleanDataPoint;
import ch.hevs.isi.core.DataPoint;
import ch.hevs.isi.core.FloatDataPoint;

public class SmartControl {

    private static final float BATTERY_LEVEL_LOW = 0.4f;
    private static final float BATTERY_LEVEL_HIGH = 0.85f;
    private static final float BATTERY_POWER_MAX = 1000f;
    private static final float COAL_AMOUNT_DAY = (1f-0.2f)/3f;
    private static final float COAL_POWER_MAX = 500f;
    private static final float FACTORY_POWER_MAX = 1000f;
    private static final float GRID_VOLTAGE_MIN = 750f;
    private static final float GRID_VOLTAGE_MAX = 900f;
    private static final float GRID_VOLTAGE_NOMINAL = 800f;
    private static final float GRID_VOLTAGE_TOLERANCE = 25f;
    private static final float PV_PANEL_NOMINAL = 1000f;
    private static final float PV_PANEL_NOMINAL_RAINING = 500f;

    public static final long TIME_BETWEEN_UPDATES = 100;



    // Useful data points
    private FloatDataPoint clock;
    private FloatDataPoint battChrg;
    private FloatDataPoint battPwrIn;
    private FloatDataPoint remoteCoalSp;
    private FloatDataPoint remoteFactorySp;
    private FloatDataPoint coalAmount;


    // Emergency data points
    private FloatDataPoint gridVoltage;

    // Ideal goal points variables
    private float igpBatteryLevel = 0f;
    private float igpCoalAmount = 0f;

    // Time variables
    private int day = 1;
    private boolean dayChanged = false;
    private float previousCoalAmount = 0f;
    private float initialTime;
    final float kpiFactory = 0.1f;
    final float kpiCoal = 0.1f;
    final float kpiCoalBatt = 0.8f;

    private void defineDataPoints(){
        do {
            try {
                Thread.sleep(20);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            clock = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("CLOCK_FLOAT");
            battChrg = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("BATT_CHRG_FLOAT");
            battPwrIn = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("BATT_P_FLOAT");
            remoteCoalSp = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("REMOTE_COAL_SP");
            remoteFactorySp = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("REMOTE_FACTORY_SP");
            coalAmount = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("COAL_AMOUNT");
            gridVoltage = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("GRID_U_FLOAT");
        } while (!testIfDataPointsAreDefined());
        System.out.println("Data points defined");
    }

    private boolean testIfDataPointsAreDefined(){
        if(clock == null) return false;
        if(battChrg == null) return false;
        if(battPwrIn == null) return false;
        if(remoteCoalSp == null) return false;
        if(remoteFactorySp == null) return false;
        if(coalAmount == null) return false;
        if(gridVoltage == null) return false;
        return true;
    }

    private float getTimeOfTheDay(){
        float timeOfTheDay = clock.getValue();
        timeOfTheDay %= 1;
        timeOfTheDay *= 24;
        return timeOfTheDay;
    }

    private float getProducerPower(){
        float producerPower = 0f;
        FloatDataPoint solar = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("SOLAR_P_FLOAT");
        FloatDataPoint wind = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("WIND_P_FLOAT");
        FloatDataPoint coal = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("COAL_P_FLOAT");

        producerPower += solar.getValue();
        producerPower += wind.getValue();
        producerPower += coal.getValue();

        return producerPower;
    }

    private float getConsumerPower(){
        float consumerPower = 0f;
        FloatDataPoint home = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("HOME_P_FLOAT");
        FloatDataPoint factory = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("FACTORY_P_FLOAT");
        FloatDataPoint bunker = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("BUNKER_P_FLOAT");

        consumerPower += home.getValue();
        consumerPower += factory.getValue();
        consumerPower += bunker.getValue();
        return consumerPower;
    }

    private boolean checkIfGridIsFine(){
        boolean gridIsFine = true;

        // Grid voltage is too high
        if(gridVoltage.getValue() > (GRID_VOLTAGE_MAX-GRID_VOLTAGE_TOLERANCE)) {
            gridIsFine = false;
            new BooleanDataPoint("SOLAR_CONNECT_ST", true).setValue(false);
            new BooleanDataPoint("WIND_CONNECT_ST", true).setValue(false);
            remoteCoalSp.setValue(0f);
            if(gridVoltage.getValue() > GRID_VOLTAGE_MAX) {
                remoteFactorySp.setValue(1f);
            }
        }

        // Grid voltage is too low
        if(gridVoltage.getValue() < (GRID_VOLTAGE_MIN+GRID_VOLTAGE_TOLERANCE)) {
            gridIsFine = false;
            new BooleanDataPoint("SOLAR_CONNECT_ST", true).setValue(true);
            new BooleanDataPoint("WIND_CONNECT_ST", true).setValue(true);
            remoteCoalSp.setValue(1f);
            if(gridVoltage.getValue() < GRID_VOLTAGE_MIN) {
                remoteFactorySp.setValue(0f);
            }
        }

        return gridIsFine;
    }

    private int getPeriodOfDay(){
        float timeOfTheDay = getTimeOfTheDay();

        // 0 = Night, 1 = Morning, 2 = Afternoon
        int periodOfDay = 0;
        if(timeOfTheDay >= 0 && timeOfTheDay < 6) periodOfDay = 1; // Night
        if(timeOfTheDay >= 6 && timeOfTheDay < 12) periodOfDay = 2; // Morning
        if(timeOfTheDay >= 12 && timeOfTheDay < 18) periodOfDay = 3; // Afternoon
        if(timeOfTheDay >= 18 && timeOfTheDay < 24) periodOfDay = 4; // Evening

        return periodOfDay;
    }

    private void defineIGPbatteryLevel(){

        switch (getPeriodOfDay()){
            case 1: // Night
                if(getTimeOfTheDay() < 5f) igpBatteryLevel = BATTERY_LEVEL_HIGH-0.5f;
                else igpBatteryLevel = BATTERY_LEVEL_LOW+0.2f;
                break;
            case 2: // Morning
                igpBatteryLevel = BATTERY_LEVEL_LOW;
                break;
            case 3: // Afternoon
                igpBatteryLevel = BATTERY_LEVEL_HIGH;
                break;
            case 4: // Evening
                if(getTimeOfTheDay() < 21f) igpBatteryLevel = BATTERY_LEVEL_HIGH-0.2f;
                else igpBatteryLevel = BATTERY_LEVEL_HIGH-0.4f;
                break;
            default:
                break;
        }
    }

    private void defineIGPcoalAmount(int day){
        float igpCoalAmountForTheDay = 1f;
        switch (getPeriodOfDay()){
            case 1: // Night
                if(getTimeOfTheDay() < 5f) igpCoalAmountForTheDay = 0.93f;
                else igpCoalAmountForTheDay = 0.8f;
                break;
            case 2: // Morning
                igpCoalAmountForTheDay = 0.7f;
                break;
            case 3: // Afternoon
                igpCoalAmountForTheDay = 0.6f;
                break;
            case 4: // Evening
                if(getTimeOfTheDay() < 21f) igpCoalAmountForTheDay = 0.5f;
                else igpCoalAmountForTheDay = 0.3f;
                break;
            default:
                break;
        }
        igpCoalAmountForTheDay += (3-day);
        igpCoalAmountForTheDay /= 3f;
        igpCoalAmount = igpCoalAmountForTheDay;


    }


    public SmartControl() {

        // start the solar and wind power plants
        new BooleanDataPoint("REMOTE_SOLAR_SW", true).setValue(true);
        new BooleanDataPoint("REMOTE_WIND_SW", true).setValue(true);
        new FloatDataPoint("REMOTE_COAL_SP", true).setValue(0f);
        new FloatDataPoint("REMOTE_FACTORY_SP", true).setValue(0f);

        defineDataPoints();

        initialTime = getTimeOfTheDay();

    }

    public boolean run() {
        float spCoal = remoteCoalSp.getValue();
        float spFactory = remoteFactorySp.getValue();

        if ((getTimeOfTheDay() < 0.5f) && !dayChanged) {
            dayChanged = true;
            day++;
        }
        if (getTimeOfTheDay() > 23f) {
            dayChanged = false;
        }
        if(coalAmount.getValue() > previousCoalAmount) {
            previousCoalAmount = coalAmount.getValue();
            day = 1;
        }

        defineIGPbatteryLevel();
        defineIGPcoalAmount(day);

        new FloatDataPoint("IGP_BATTERY_LEVEL", true).setValue(igpBatteryLevel);
        new FloatDataPoint("IGP_COAL_AMOUNT", true).setValue(igpCoalAmount);

        String debug = "Day " + day + " - ";
        //debug += "IGP battery level: " + igpBatteryLevel;
        //debug += " IGP coal amount: " + igpCoalAmount;
        //  debug += " - ";


        float battDiff = battChrg.getValue() - igpBatteryLevel;
        new FloatDataPoint("BATT_DIFF", true).setValue(battDiff);
        debug += "Batt. diff.: " + battDiff + " - ";

        float igpEnergyBattery = battDiff * BATTERY_POWER_MAX / 0.4f; // Over energy in battery
        new FloatDataPoint("IGP_ENERGY_BATTERY", true).setValue(igpEnergyBattery);
        debug += "IGP en. batt.: " + igpEnergyBattery + " - ";

        float powerDiff = getProducerPower() - getConsumerPower();
        new FloatDataPoint("POWER_DIFF", true).setValue(powerDiff);
        debug += "Power diff: " + powerDiff + " - ";

        float balance = powerDiff + igpEnergyBattery; // Energy to use
        new FloatDataPoint("BALANCE", true).setValue(balance);
        debug += "Balance: " + balance + " - ";

        float coalDiff = coalAmount.getValue() - igpCoalAmount; // Over coal in storage
        new FloatDataPoint("COAL_DIFF", true).setValue(coalDiff);
        //debug += "Coal diff.: " + coalDiff + " - ";

        if(balance < 0){ // Produce more energy or consume less
            if(coalDiff > 0){ // Produce more energy
                spCoal += kpiCoal * (balance / COAL_POWER_MAX);
                spFactory -= kpiFactory * (balance / FACTORY_POWER_MAX);
                debug += "Produce more energy";
            } else { // Consume less energy
                spFactory -= kpiFactory * (balance / FACTORY_POWER_MAX);
                spCoal += kpiCoal * (balance / COAL_POWER_MAX);
                debug += "Consume less energy";
            }
        } else { // Produce less energy or consume more
            if(coalDiff < 0){ // Produce less energy
                spCoal -= kpiCoal * (balance / COAL_POWER_MAX);
                spFactory += kpiFactory * (balance / FACTORY_POWER_MAX);
                debug += "Produce less energy";
            } else { // Consume more energy
                spFactory += kpiFactory * (balance / FACTORY_POWER_MAX);
                spCoal -= kpiCoal * (balance / COAL_POWER_MAX);
                debug += "Consume more energy";
            }
        }

        FloatDataPoint solar = (FloatDataPoint) DataPoint.getDataPointOnListFromLabel("SOLAR_P_FLOAT");
        float foo = (1000f- solar.getValue()) / 1000f;

        if(spCoal > foo) spCoal = foo;





        if (spFactory > 1f) spFactory = 1f;
        if (spFactory < 0f) spFactory = 0f;
        if (spCoal > 1f) spCoal = 1f;
        if (spCoal < 0f) spCoal = 0f;

        new FloatDataPoint("REMOTE_FACTORY_SP", true).setValue(spFactory);
        new FloatDataPoint("REMOTE_COAL_SP", true).setValue(spCoal);

        System.out.println(debug);

        return checkIfGridIsFine();
    }
}
